// Copyright (c) 2017, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by a MIT-style license
// that can be found in the LICENSE file.

part of editex;

const _cursorCommand = r'\phantom{~}';
const _emptyCommand = r'\phantom{1}';

class EdiTeXCommand {
  final String name, latexTemplate, parseTemplate;
  const EdiTeXCommand(this.name, this.latexTemplate, this.parseTemplate);
}

/// Interface between editor and environment. When the user enters a command
/// this function is called to find a command template. If null is returned, the
/// command is considered unknown.
typedef EdiTeXCommand CommandResolverFn(String command);

/// Same as [CommandResolverFn] but for instant commands.
typedef EdiTeXCommand InstantCommandResolverFn(String command);

/// Command used when the user types '/'. This cannot be configured. It is
/// considered a good practice to stick to LaTeX like commands when those are
/// present.
const _divideCommand = 'frac';

const _defaultCommands = const {
  'a': const EdiTeXCommand('a', r'{}_\text{?}a', 'a'),
  'frac': const EdiTeXCommand('frac', r'\frac{\,$0\,}{\,$1\,}', r'frac($0,$1)'),
  'sqrt': const EdiTeXCommand('sqrt', r'\sqrt{$0\,}', r'sqrt($0)'),
  'root': const EdiTeXCommand('root', r'\sqrt[$0]{$1\,}', r'root($0,$1)'),
  'log': const EdiTeXCommand('log', r'\log_{$0}$1', r'log($0,$1)'),
  'diff': const EdiTeXCommand(
      'diff', r'\frac{\partial}{\partial $0}$1', r'diff($0,$1)'),
  'lim': const EdiTeXCommand(
      'lim', r'~\lim\limits_{$0\to $1}$2', r'lim($0,$1,$2)'),
  'int': const EdiTeXCommand(
      'int', r'~\displaystyle\int_{$0}^{$1}$2~\text{d}$3', r'int($3,$0,$1,$2)')
};

const _defaultInstantCommands = const {
  '(': const EdiTeXCommand('(', r'\left($0\right)', r'($0)'),
  '[': const EdiTeXCommand('(', r'\left[$0\right]', r'brackets($0)'),
  '^': const EdiTeXCommand('^', r'^{$0}', r'^($0)'),
  '_': const EdiTeXCommand('_', r'_{$0}', r'_($0)'),
  '*': const EdiTeXCommand('*', r'{}\cdot{}', '*'),
  '+': const EdiTeXCommand('+', r'{}+{}', '+'),
  '-': const EdiTeXCommand('-', r'{}-{}', '-'),
  '~': const EdiTeXCommand('~', r'-', '-'),
  '!': const EdiTeXCommand('!', r'!', '!')
};

const _subSupCommand =
    const EdiTeXCommand('subsup', r'_{$0}^{$1}', r'_($0)^($1)');

EdiTeXCommand _defaultCommandResolver(String command) {
  return _defaultCommands[command];
}

EdiTeXCommand _defaultInstantCommandResolver(String command) {
  return _defaultInstantCommands[command];
}

/// Interface functions.
abstract class EdiTeXInterface {
  EdiTeXCommand resolveCommand(String command);
  EdiTeXCommand resolveInstantCommand(String command);
  bool hasCommand(String command);
  bool hasInstantCommand(String command);
}

/// Static function implementation of [EdiTeXInterface].
class StatecEdiTeXInterface implements EdiTeXInterface {
  final CommandResolverFn _resolveCommand;
  final InstantCommandResolverFn _resolveInstantCommand;

  const StatecEdiTeXInterface(
      this._resolveCommand, this._resolveInstantCommand);

  @override
  EdiTeXCommand resolveCommand(command) => _resolveCommand(command);

  @override
  EdiTeXCommand resolveInstantCommand(command) =>
      _resolveInstantCommand(command);

  @override
  bool hasCommand(String command) {
    final template = resolveCommand(command);
    return template != null;
  }

  @override
  bool hasInstantCommand(String command) {
    final template = resolveInstantCommand(command);
    return template != null;
  }
}

/// For [EdiTeX.undoHistory] and [EdiTeX.redoHistory].
class _EditHistoryEntry {
  final Fragment content;
  final int cursorIdx;
  _EditHistoryEntry(this.content, this.cursorIdx);
}

/// EdiTeX editor component.
/// Can initialize itself in an empty <div>
class EdiTeX {
  final EdiTeXInterface interface;
  final DivElement container, katexWrapper;

  final undoHistory = new List<_EditHistoryEntry>();
  final redoHistory = new List<_EditHistoryEntry>();
  var content = new MultiFragment();

  var cursorIdx = 0;
  var doUpdate = true;

  final _onLeftLeave = new StreamController<Null>.broadcast();
  final _onRightLeave = new StreamController<Null>.broadcast();

  EdiTeX(this.container,
      [this.interface = const StatecEdiTeXInterface(
          _defaultCommandResolver, _defaultInstantCommandResolver)])
      : katexWrapper = new DivElement() {
    container.append(katexWrapper);

    // Make focussable via keyboard.
    container.tabIndex = 0;

    // Add keyboard listeners.
    container.onKeyDown.listen(onKeyDown);

    // Finish command on unfocus.
    container.onBlur.listen((_) {
      _finishCurrentCommand(true);
    });

    new Timer.periodic(new Duration(milliseconds: 30), (t) {
      if (doUpdate) {
        // Set [doUpdate] to false first in case it is updated again while
        // rendering.
        doUpdate = false;

        // Render content using KaTeX.
        var renderedSuccesfully = false;
        while (!renderedSuccesfully) {
          try {
            katex.render(getLaTeX(), katexWrapper,
                new katex.RenderingOptions(displayMode: true));
            renderedSuccesfully = true;
          } catch (e) {
            // Destructive undo
            undo(true);
          }
        }

        // Find cursor.
        final cursor = katexWrapper
            .querySelector("span.mspace[style='color: transparent;']");
        cursor.classes.add('editex-cursor');
        final katexDisplay = katexWrapper.children.first;

        // Compute katexDiv offset using cursor position
        var cursorOffset = cursor.getBoundingClientRect().left;
        final wrapperWidth = katexWrapper.getBoundingClientRect().width;
        final katexRect = katexDisplay.getBoundingClientRect();
        cursorOffset -= katexRect.left - 1; // -1 because of cursor margin.

        if (cursorOffset > wrapperWidth) {
          katexDisplay.style.marginLeft = '-${cursorOffset - wrapperWidth}px';
        } else {
          katexDisplay.style.marginLeft = '0px';
        }
      }
    });
  }

  Future destroy() {
    return Future.wait([_onLeftLeave.close(), _onRightLeave.close()]);
  }

  Stream<Null> get onLeftLeave => _onLeftLeave.stream;
  Stream<Null> get onRightLeave => _onRightLeave.stream;

  bool get isEmpty => content.fragments.isEmpty;
  bool get isNotEmpty => content.fragments.isNotEmpty;

  /// Add undo checkpoint.
  void _addUndoCheckpoint() {
    undoHistory.add(new _EditHistoryEntry(content.clone(), cursorIdx));
    redoHistory.clear();
  }

  /// Undo last change.
  void undo([bool destructive = false]) {
    if (undoHistory.isNotEmpty) {
      if (!destructive) {
        redoHistory.add(new _EditHistoryEntry(content.clone(), cursorIdx));
      }

      final replace = undoHistory.removeLast();
      content = replace.content;
      cursorIdx = replace.cursorIdx;
      doUpdate = true;
    }
  }

  /// Redo last change.
  void redo() {
    if (redoHistory.isNotEmpty) {
      undoHistory.add(new _EditHistoryEntry(content.clone(), cursorIdx));
      final replace = redoHistory.removeLast();
      content = replace.content;
      cursorIdx = replace.cursorIdx;
      doUpdate = true;
    }
  }

  /// Keyboard handler.
  void onKeyDown(KeyboardEvent event) {
    // Keystrokes
    if (event.ctrlKey) {
      if (event.keyCode == KeyCode.Z) {
        if (event.shiftKey) {
          redo();
        } else {
          undo();
        }
      }
    }

    // Normal insert
    else if (new RegExp(r'^[0-9A-Za-z]$').hasMatch(event.key)) {
      _addUndoCheckpoint();

      var char = event.key;
      if (event.shiftKey) {
        char = char.toUpperCase();
      }

      cursorIdx += content
          .insert(new W<int>(cursorIdx), new CharFragment(char), interface)
          .cursorOffset;
    }

    // Insert backward slash (command separator).
    else if (event.key == r'\') {
      _addUndoCheckpoint();
      _finishCurrentCommand();

      cursorIdx += content
          .insert(new W<int>(cursorIdx), new CharFragment(event.key), interface)
          .cursorOffset;
    }

    // Instant command insert
    else if (interface.hasInstantCommand(event.key)) {
      _addUndoCheckpoint();
      _finishCurrentCommand();

      final fragment =
          new CommandFragment(interface.resolveInstantCommand(event.key));
      cursorIdx += content
          .insert(new W<int>(cursorIdx), fragment, interface)
          .cursorOffset;
    }

    // Special divide behavior
    else if (event.key == '/') {
      // Only proceed if [_divideCommand] is specified.
      if (interface.hasCommand(_divideCommand)) {
        _addUndoCheckpoint();
        _finishCurrentCommand();

        cursorIdx += content
            .insert(new W<int>(cursorIdx), new CharFragment('/'), interface)
            .cursorOffset;
      }
    }

    // Finish a command.
    else if (event.keyCode == KeyCode.SPACE) {
      _finishCurrentCommand(true);
    }

    // Remove before cursor.
    else if (event.keyCode == KeyCode.BACKSPACE) {
      _addUndoCheckpoint();
      final result = content.remove(new W<int>(cursorIdx), -1);
      cursorIdx -= result.count;
    }

    // Remove after cursor.
    else if (event.keyCode == KeyCode.DELETE) {
      final result = content.remove(new W<int>(cursorIdx), 0);
      cursorIdx -= result.removeSelf ? result.count : 0;
    }

    // Arrow key navigation.
    else if (event.keyCode == KeyCode.LEFT) {
      _finishCurrentCommand(true);
      cursorIdx--;
      if (cursorIdx == -1) {
        cursorIdx = 0;
        _onLeftLeave.add(null);
        return;
      }
    } else if (event.keyCode == KeyCode.RIGHT) {
      _finishCurrentCommand(true);
      cursorIdx++;
      if (cursorIdx == content.length) {
        cursorIdx--;
        _onRightLeave.add(null);
        return;
      }
    } else if (event.keyCode == KeyCode.HOME) {
      _finishCurrentCommand(true);
      cursorIdx = 0;
    } else if (event.keyCode == KeyCode.END) {
      _finishCurrentCommand(true);
      cursorIdx = content.length - 1;
    }

    // Catch other cases to prevent a KaTeX update for nothing.
    else {
      return;
    }

    // Something was changed, so prevent default event response.
    event.preventDefault();

    doUpdate = true;
  }

  void _finishCurrentCommand([bool addUndoCheckpoint = false]) {
    // Cannot use [_addUndoCheckpoint] because the content from before finishing
    // must be added to the history.
    final contentBefore = addUndoCheckpoint ? content.clone() : null;
    final result = content.finish(new W<int>(cursorIdx), interface);
    if (result < 0) {
      if (addUndoCheckpoint) {
        undoHistory.add(new _EditHistoryEntry(contentBefore, cursorIdx));
        redoHistory.clear();
      }
      cursorIdx += result + 1;
      doUpdate = true;
    }
  }

  /// Get JSON serializable data object.
  List getData() => content.getData();

  /// Load data generated with [getData].
  void loadData(List data) {
    _addUndoCheckpoint();
    content = new MultiFragment()..loadData(data, interface);
    cursorIdx = 0;
    doUpdate = true;
  }

  /// Get content as parsable expression.
  String getParsableContent() => content.getParsableContent().trim();

  /// Get LaTeX string with cursor.
  String getLaTeX() => content.render(new W<int>(cursorIdx));
}

class W<T> {
  T v;
  W(this.v);
}
